/*
 * Copyright © 2011 Linaro Limited
 *
 * This file is part of glcompbench.
 *
 * glcompbench is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * glcompbench is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with glcompbench.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authors:
 *  Alexandros Frantzis <alexandros.frantzis@linaro.org>
 *  Jesse Barker <jesse.barker@linaro.org>
 */

#include "benchmark.h"
#include "log.h"
#include <sstream>

using std::string;
using std::vector;
using std::map;

std::map<string, CompositeTest *> Benchmark::test_map_;

static void
split(const string &s, char delim, vector<string> &elems)
{
    std::stringstream ss(s);

    string item;
    while(std::getline(ss, item, delim))
        elems.push_back(item);
}

static CompositeTest &
get_test_from_description(const string &s)
{
    vector<string> elems;

    split(s, ':', elems);

    const string &name = !elems.empty() ? elems[0] : ""; 

    return Benchmark::get_test_by_name(name);
}

static vector<Benchmark::OptionPair>
get_options_from_description(const string &s)
{
    vector<Benchmark::OptionPair> options;
    vector<string> elems;

    split(s, ':', elems);

    for (vector<string>::const_iterator iter = ++elems.begin();
         iter != elems.end();
         iter++)
    {
        vector<string> opt;

        split(*iter, '=', opt);
        if (opt.size() == 2)
            options.push_back(Benchmark::OptionPair(opt[0], opt[1]));
        else
            Log::info("Warning: ignoring invalid option string '%s' "
                      "in benchmark description\n",
                      iter->c_str());
    }

    return options;
}

void
Benchmark::register_test(CompositeTest &test)
{
    test_map_[test.name()] = &test;
}

CompositeTest &
Benchmark::get_test_by_name(const string &name)
{
    map<string, CompositeTest *>::const_iterator iter;

    if ((iter = test_map_.find(name)) != test_map_.end())
        return *(iter->second);
    else
        return CompositeTest::dummy();
}

Benchmark::Benchmark(CompositeTest &test, const vector<OptionPair> &options) :
    test_(test), options_(options)
{
}

Benchmark::Benchmark(const string &name, const vector<OptionPair> &options) :
    test_(Benchmark::get_test_by_name(name)), options_(options)
{
}

Benchmark::Benchmark(const string &s) :
    test_(get_test_from_description(s)),
    options_(get_options_from_description(s))
{
}

CompositeTest &
Benchmark::setup_test()
{
    test_.reset_options();
    load_options();

    test_.prepare_for_run();

    return test_;
}

void
Benchmark::teardown_test()
{
    test_.cleanup();
}

void
Benchmark::load_options()
{
    for (vector<OptionPair>::iterator iter = options_.begin();
         iter != options_.end();
         iter++)
    {
        test_.set_option(iter->first, iter->second);
    }
}

