/*
 * Copyright © 2011 Linaro Limited
 *
 * This file is part of glcompbench.
 *
 * glcompbench is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * glcompbench is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with glcompbench.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authors:
 *  Alexandros Frantzis <alexandros.frantzis@linaro.org>
 *  Jesse Barker <jesse.barker@linaro.org>
 */

#include <string>
#include <sstream>

#include "gl-headers.h"

#include <EGL/egl.h>
#include <EGL/eglext.h>

#include "composite-canvas-egl.h"
#include "composite-window-eglimage.h"
#include "composite-window-ximage.h"
#include "options.h"
#include "log.h"

PFNEGLCREATEIMAGEKHRPROC eglCreateImageKHR_;
PFNEGLDESTROYIMAGEKHRPROC eglDestroyImageKHR_;
PFNGLEGLIMAGETARGETTEXTURE2DOESPROC glEGLImageTargetTexture2DOES_;

bool
CompositeCanvasEGL::check_eglimage_extensions()
{
    /*
     * Parse the extensions we care about from the extension string.
     * Don't even bother to get function pointers until we know the
     * extension is present.
     */
    bool have_exts(true);
    std::string extString;
    const char* exts = eglQueryString(egl_display_, EGL_EXTENSIONS);
    if (exts) {
        extString = exts;
    }

    Log::debug("==== EGL extensions ====\n%s\n", extString.c_str());
    if (extString.find("EGL_KHR_image_pixmap") != std::string::npos) {
        eglCreateImageKHR_ =
            reinterpret_cast<PFNEGLCREATEIMAGEKHRPROC>(eglGetProcAddress("eglCreateImageKHR"));
        eglDestroyImageKHR_ =
            reinterpret_cast<PFNEGLDESTROYIMAGEKHRPROC>(eglGetProcAddress("eglDestroyImageKHR"));
    }

    if (!eglCreateImageKHR_ || !eglDestroyImageKHR_)
    {
        Log::info("** EGL does not support EGL_KHR_image_pixmap!\n");
        have_exts = false;
    }

    exts = reinterpret_cast<const char*>(glGetString(GL_EXTENSIONS));
    if (exts) {
        extString = exts;
    }

    Log::debug("==== GL extensions ====\n%s\n", extString.c_str());
    if (extString.find("GL_OES_EGL_image") != std::string::npos) {
        glEGLImageTargetTexture2DOES_ = reinterpret_cast<PFNGLEGLIMAGETARGETTEXTURE2DOESPROC>(
            eglGetProcAddress("glEGLImageTargetTexture2DOES"));
    }
    if (!glEGLImageTargetTexture2DOES_)
    {
        Log::info("** OpenGL ES2.0 does not support GL_OES_EGL_image!\n");
        have_exts = false;
    }
    Log::debug("=======================\n");

    return have_exts;
}

bool
CompositeCanvasEGL::ensure_egl_display()
{
    if (egl_display_)
        return true;

    egl_display_ = eglGetDisplay((EGLNativeDisplayType) xdpy_);
    if (!egl_display_) {
        Log::error("Error: eglGetDisplay() failed with error: %d\n",
                     eglGetError());
        return false;
    }
    if (!eglInitialize(egl_display_, NULL, NULL)) {
        Log::error("Error: eglInitialize() failed with error: %d\n",
                     eglGetError());
        return false;
        egl_display_ = 0;
    }

    return true;
}

bool
CompositeCanvasEGL::ensure_egl_config()
{
    static const EGLint attribs[] = {
        EGL_RED_SIZE, 1,
        EGL_GREEN_SIZE, 1,
        EGL_BLUE_SIZE, 1,
        EGL_ALPHA_SIZE, 1,
        EGL_DEPTH_SIZE, 1,
#ifdef USE_GLES2
        EGL_RENDERABLE_TYPE, EGL_OPENGL_ES2_BIT,
#elif USE_GL
        EGL_RENDERABLE_TYPE, EGL_OPENGL_BIT,
#endif
        EGL_NONE
    };
    EGLint num_configs;
    EGLint vid;

    if (egl_config_)
        return true;

    ensure_egl_display();

    if (!eglChooseConfig(egl_display_, attribs, &egl_config_,
                         1, &num_configs))
    {
        Log::error("Error: eglChooseConfig() failed with error: %d\n",
                     eglGetError());
        return false;
    }

    for (int i = 0; attribs[i] != EGL_NONE; i += 2) {
        if (eglGetConfigAttrib(egl_display_, egl_config_, attribs[i], &vid))
            Log::debug("%d: %d\n", i, vid);
    }

    if (!eglGetConfigAttrib(egl_display_, egl_config_,
                            EGL_NATIVE_VISUAL_ID, &vid))
    {
        Log::error("Error: eglGetConfigAttrib() failed with error: %d\n",
                     eglGetError());
        return false;
    }

    return true;
}

bool
CompositeCanvasEGL::ensure_egl_surface()
{
    static const EGLint ctx_attribs[] = {
#ifdef USE_GLES2
        EGL_CONTEXT_CLIENT_VERSION, 2,
#endif
        EGL_NONE
    };

    if (egl_surface_)
        return true;

    if (!ensure_egl_display())
        return false;

#ifdef USE_GLES2
    eglBindAPI(EGL_OPENGL_ES_API);
#elif USE_GL
    eglBindAPI(EGL_OPENGL_API);
#endif

    egl_context_ = eglCreateContext(egl_display_, egl_config_,
                                    EGL_NO_CONTEXT, ctx_attribs);
    if (!egl_context_) {
        Log::error("Error: eglCreateContext() failed with error: %d\n",
                     eglGetError());
        return false;
    }

    egl_surface_ = eglCreateWindowSurface(egl_display_, egl_config_,
                                          (EGLNativeWindowType) canvas_,
                                          NULL);
    if (!egl_surface_) {
        Log::error("Error: eglCreateWindowSurface failed with error: %d\n",
                     eglGetError());
        return false;
    }

    return true;
}

XVisualInfo *
CompositeCanvasEGL::get_canvas_xvisualinfo()
{
    XVisualInfo vis_tmpl;
    XVisualInfo *vis_info;
    int num_visuals;
    EGLint vid;

    if (!ensure_egl_config())
        return 0;

    if (!eglGetConfigAttrib(egl_display_, egl_config_,
                            EGL_NATIVE_VISUAL_ID, &vid))
    {
        Log::error("Error: eglGetConfigAttrib() failed with error: %d\n",
                     eglGetError());
        return 0;
    }

    /* The X window visual must match the EGL config */
    vis_tmpl.visualid = vid;
    vis_info = XGetVisualInfo(xdpy_, VisualIDMask, &vis_tmpl,
                             &num_visuals);
    if (!vis_info) {
        Log::error("Error: couldn't get X visual\n");
        return 0;
    }

    return vis_info;
}

bool
CompositeCanvasEGL::make_current()
{
    if (!ensure_egl_surface())
        return false;

    if (!eglMakeCurrent(egl_display_, egl_surface_, egl_surface_, egl_context_)) {
        Log::error("Error: eglMakeCurrent failed with error %d\n", eglGetError());
        return false;
    }

    if (!eglSwapInterval(egl_display_, 0))
        Log::info("** Failed to set swap interval. Results may be bounded above by refresh rate.\n");

    use_accel_tfp_ = Options::use_accel_tfp;

    if (Options::use_accel_tfp && !check_eglimage_extensions()) {
        use_accel_tfp_ = false;
        Log::info("** Required EGL/GLES2.0 extensions not present.\n"
                       "** Falling back to XImage/glTexImage2D texture update method!\n");
    }

    return true;
}


CompositeWindow *
CompositeCanvasEGL::create_window(Window win)
{
    if (use_accel_tfp_)
        return new CompositeWindowEGLImage(xdpy_, win, egl_display_);
    else
        return new CompositeWindowXImage(xdpy_, win);
}

void
CompositeCanvasEGL::update_screen()
{
    if (Options::swap_buffers)
        eglSwapBuffers(egl_display_, egl_surface_);
    else
        glFinish();
}

std::string
CompositeCanvasEGL::info_string()
{
    make_current();

    std::stringstream ss;

    ss << "    OpenGL Information" << std::endl;
    ss << "    GL_VENDOR:     " << glGetString(GL_VENDOR) << std::endl;
    ss << "    GL_RENDERER:   " << glGetString(GL_RENDERER) << std::endl;
    ss << "    GL_VERSION:    " << glGetString(GL_VERSION) << std::endl;

    return ss.str();
}
