/*
 * Copyright © 2011 Linaro Limited
 *
 * This file is part of glcompbench.
 *
 * glcompbench is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * glcompbench is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with glcompbench.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authors:
 *  Alexandros Frantzis <alexandros.frantzis@linaro.org>
 *  Jesse Barker <jesse.barker@linaro.org>
 */

#include <string>
#include <sstream>

#include <GL/gl.h>
#include <GL/glx.h>

#include "composite-canvas-glx.h"
#include "composite-window-glxpixmap.h"
#include "composite-window-ximage.h"
#include "options.h"
#include "log.h"

PFNGLXBINDTEXIMAGEEXTPROC glXBindTexImageEXT_;
PFNGLXSWAPINTERVALEXTPROC glXSwapIntervalEXT_;
PFNGLXSWAPINTERVALMESAPROC glXSwapIntervalMESA_;

bool
CompositeCanvasGLX::check_glx_version()
{
    int glx_major, glx_minor;

    if (!glXQueryVersion(xdpy_, &glx_major, &glx_minor ) ||
        (glx_major == 1  && glx_minor < 3) || glx_major < 1)
    {
        Log::error("GLX version >= 1.3 is required\n");
        return false;
    }

    return true;
}

void
CompositeCanvasGLX::init_extensions()
{
    /*
     * Parse the extensions we care about from the extension string.
     * Don't even bother to get function pointers until we know the
     * extension is present.
     */
    std::string extString;
    const char* exts = glXQueryExtensionsString(xdpy_, 0);
    if (exts) {
        extString = exts;
    }

    Log::debug("==== GLX extensions ====\n%s\n", extString.c_str());

    /* GLX_EXT_texture_from_pixmap */
    if (extString.find("GLX_EXT_texture_from_pixmap") != std::string::npos) {
        glXBindTexImageEXT_ =
            reinterpret_cast<PFNGLXBINDTEXIMAGEEXTPROC>(
                glXGetProcAddress((const GLubyte *)"glXBindTexImageEXT"));
    }

    if (!glXBindTexImageEXT_) {
        Log::info("** GLX does not support GLX_EXT_texture_from_pixmap!\n");
    }

    /* 
     * GLX_EXT_swap_control or GL_MESA_swap_control. Note that
     * GLX_SGI_swap_control is not enough because it doesn't allow 0 as a valid
     * value (i.e. you can't turn off VSync).
     */
    if (extString.find("GLX_EXT_swap_control") != std::string::npos) {
        glXSwapIntervalEXT_ =
            reinterpret_cast<PFNGLXSWAPINTERVALEXTPROC>(
                glXGetProcAddress((const GLubyte *)"glXSwapIntervalEXT"));
    }
    else if (extString.find("GLX_MESA_swap_control") != std::string::npos) {
        glXSwapIntervalMESA_ =
            reinterpret_cast<PFNGLXSWAPINTERVALMESAPROC>(
                glXGetProcAddress((const GLubyte *)"glXSwapIntervalMESA"));
    }


    if (!glXSwapIntervalEXT_ && !glXSwapIntervalMESA_) {
        Log::info("** GLX does not support GLX_EXT_swap_control or GLX_MESA_swap_control!\n");
    }

    Log::debug("=======================\n");
}

bool
CompositeCanvasGLX::ensure_glx_fbconfig()
{
    static int attribs[] = {
        GLX_X_RENDERABLE, True,
        GLX_DRAWABLE_TYPE, GLX_WINDOW_BIT,
        GLX_RENDER_TYPE, GLX_RGBA_BIT,
        GLX_X_VISUAL_TYPE, GLX_TRUE_COLOR,
        GLX_RED_SIZE, 1,
        GLX_GREEN_SIZE, 1,
        GLX_BLUE_SIZE, 1,
        GLX_ALPHA_SIZE, 1,
        GLX_DEPTH_SIZE, 1,
        GLX_DOUBLEBUFFER, True,
        None
    };
    int num_configs;

    if (!check_glx_version())
        return false;

    GLXFBConfig *fbc = glXChooseFBConfig(xdpy_, DefaultScreen(xdpy_),
                                         attribs, &num_configs);
    if (!fbc) {
        Log::error("Error: glXChooseFBConfig() failed\n");
        return false;
    }

    Log::debug("Found %d matching FB configs.\n", num_configs);

    /* Get the first matching config */
    glx_fbconfig_ = fbc[0];

    XFree(fbc);

    return true;
}

bool
CompositeCanvasGLX::ensure_glx_context()
{
    if (glx_context_)
        return true;

    if (!ensure_glx_fbconfig())
        return false;

    glx_context_ = glXCreateNewContext(xdpy_, glx_fbconfig_, GLX_RGBA_TYPE,
                                       0, True);
    if (!glx_context_) {
        Log::error("Error: glXCreateNewContext failed\n");
        return false;
    }

    return true;
}

XVisualInfo *
CompositeCanvasGLX::get_canvas_xvisualinfo()
{
    if (!ensure_glx_fbconfig())
        return 0;

    XVisualInfo *vis_info = glXGetVisualFromFBConfig(xdpy_, glx_fbconfig_ );

    return vis_info;
}

bool
CompositeCanvasGLX::make_current()
{
    if (!ensure_glx_context())
        return false;

    init_extensions();

    if (!glXMakeCurrent(xdpy_, canvas_, glx_context_)) {
        Log::error("Error: glXMakeCurrent failed\n");
        return false;
    }

    if ((!glXSwapIntervalEXT_ || glXSwapIntervalEXT_(xdpy_, canvas_, 0)) &&
        (!glXSwapIntervalMESA_ || glXSwapIntervalMESA_(0)))
    {
        Log::info("** Failed to set swap interval. Results may be bounded above by refresh rate.\n");
    }

    use_accel_tfp_ = Options::use_accel_tfp;

    if (Options::use_accel_tfp && !glXBindTexImageEXT_) {
        use_accel_tfp_ = false;
        Log::info("** Required GLX extensions not present.\n"
                  "** Falling back to XImage/glTexImage2D texture update method!\n");
    }

    return true;
}


CompositeWindow *
CompositeCanvasGLX::create_window(Window win)
{
    if (use_accel_tfp_)
        return new CompositeWindowGLXPixmap(xdpy_, win, glx_fbconfig_);
    else
        return new CompositeWindowXImage(xdpy_, win);
}

void
CompositeCanvasGLX::update_screen()
{
    if (Options::swap_buffers)
        glXSwapBuffers(xdpy_, canvas_);
    else
        glFinish();
}

std::string
CompositeCanvasGLX::info_string()
{
    make_current();

    std::stringstream ss;

    ss << "    OpenGL Information" << std::endl;
    ss << "    GL_VENDOR:     " << glGetString(GL_VENDOR) << std::endl;
    ss << "    GL_RENDERER:   " << glGetString(GL_RENDERER) << std::endl;
    ss << "    GL_VERSION:    " << glGetString(GL_VERSION) << std::endl;

    return ss.str();
}
