/*
 * Copyright © 2011 Linaro Limited
 *
 * This file is part of glcompbench.
 *
 * glcompbench is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * glcompbench is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with glcompbench.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authors:
 *  Alexandros Frantzis <alexandros.frantzis@linaro.org>
 *  Jesse Barker <jesse.barker@linaro.org>
 */

#define __STDC_FORMAT_MACROS 1
#include <stdint.h>
#include <inttypes.h>
#include <sys/time.h>
#include <sstream>
#include <string>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/extensions/Xcomposite.h>
#include <X11/extensions/Xdamage.h>

#include "composite-canvas.h"
#include "composite-window.h"
#include "benchmark.h"
#include "options.h"
#include "log.h"
#include "profiler.h"

const std::string glcompbench_version_string("glcompbench "GLCOMPBENCH_VERSION);

/*******************
 * Private methods *
 *******************/

/**
 * Finds the CompositeWindow associated with a Window XID.
 *
 * @param win the Window XID to look for
 *
 * @return the CompositeWindow found or NULL
 */
CompositeWindow *
CompositeCanvas::find_window(Window win)
{
    for(std::list<CompositeWindow*>::iterator iter = window_list_.begin();
        iter != window_list_.end(); ++iter)
    {
        if (win == (*iter)->get_xwindow())
            return *iter;
    }

    return 0;
}

/**
 * Adds a window to list of tracked windows.
 *
 * @param win the Window XID
 */
void
CompositeCanvas::add_window(Window win)
{
    if (win == canvas_)
        return;

    char *win_name(0);
    XFetchName(xdpy_, win, &win_name);

    Log::debug("Composite: Add Window ID: 0x%x Name: %s\n",
              win, win_name);

    XFree(win_name);

    /* No point handling input-only windows */
    XWindowAttributes attr;
    XGetWindowAttributes (xdpy_, win, &attr);
    if (attr.c_class == InputOnly) {
        Log::debug("    => is input-only... ignoring\n");
        return;
    }

    /* Make sure we haven't already added the window */
    CompositeWindow *comp_win = find_window(win);
    if (comp_win) {
        Log::debug("    => already in list!\n",
                  (unsigned)win);
        return;
    }

    /* Create the CompositeWindow */
    comp_win = create_window(win);
    XSelectInput(xdpy_, win, StructureNotifyMask);

    /* Add it to the list */
    window_list_.push_back(comp_win);

    /* Update the window */
    comp_win->update();
}

/**
 * Removes a window from the list of tracked windows.
 *
 * @param win the Window XID
 */
void
CompositeCanvas::remove_window(Window win)
{
    for(std::list<CompositeWindow*>::iterator iter = window_list_.begin();
        iter != window_list_.end(); ++iter)
    {
        if (win == (*iter)->get_xwindow()) {
            delete *iter;
            window_list_.erase(iter);
            break;
        }
    }

    return;
}

/**
 * Updates a window.
 *
 * @param win the Window XID
 */
void
CompositeCanvas::update_window(Window win)
{
    /* find window in window list */
    CompositeWindow *comp_win = find_window(win);

    if (!comp_win) {
        Log::debug("Cannot find window 0x%x in list!\n",
                  (unsigned) win);
        return;
    }

    comp_win->update();
}

/**
 * Reshapes the canvas.
 *
 * @param width the new canvas width (in pixels)
 * @param height the new canvas size (in pixels)
 */
void
CompositeCanvas::reshape(int width, int height)
{
    if (current_test_)
        current_test_->reshape(width, height);
}

/**
 * Creates the X window that will contain the canvas.
 *
 * The window position and size are only hints and may not be honored
 * by the windowing system.
 *
 * @param name the window name
 * @param x the x position of the window
 * @param y the y position of the window
 * @param width the width of the window
 * @param height the height of the window
 * @param vis_info XVisualInfo used to create the window with
 *
 * @return the create window XID
 */
Window
CompositeCanvas::create_canvas_x_window(const char *name, int x, int y,
                                        int width, int height,
                                        XVisualInfo *vis_info)
{
    XSetWindowAttributes attr;
    unsigned long mask;
    Window win;

    /* window attributes */
    attr.background_pixel = 0;
    attr.border_pixel = 0;
    attr.colormap = XCreateColormap(xdpy_, root_, vis_info->visual, AllocNone);
    attr.event_mask = StructureNotifyMask | ExposureMask;
    mask = CWBackPixel | CWBorderPixel | CWColormap | CWEventMask;

    win = XCreateWindow(xdpy_, root_, 0, 0, width, height,
                        0, vis_info->depth, InputOutput,
                        vis_info->visual, mask, &attr);

    /* set hints and properties */
    {
        XSizeHints sizehints;
        sizehints.x = x;
        sizehints.y = y;
        sizehints.width  = width;
        sizehints.height = height;
        sizehints.flags = USSize | USPosition;
        XSetNormalHints(xdpy_, win, &sizehints);
        XSetStandardProperties(xdpy_, win, name, name,
                               None, NULL, 0, &sizehints);
    }

    return win;
}

/**
 * Checks if the extensions required for compositing exist.
 *
 * @return whether the required extensions exist
 */
bool
CompositeCanvas::check_extensions()
{
    int composite_event, composite_opcode, error;
    if (!XQueryExtension(xdpy_, COMPOSITE_NAME, &composite_opcode,
                         &composite_event, &error))
    {
        Log::error("No composite extension\n");
        return false;
    }

    int composite_major, composite_minor;
    XCompositeQueryVersion(xdpy_, &composite_major, &composite_minor);
    if (composite_minor < 2) {
        Log::error("Composite extension < 0.2\n");
        return false;
    }

    if (!XDamageQueryExtension(xdpy_, &damage_event_, &error))
    {
        Log::error("No damage extension\n");
        return false;
    }

    return true;
}

/**
 * Redirects the window drawing to offscreen pixmaps using
 * the XComposite extension.
 *
 * The list of windows to redirect is sby the Options::window_ids option.
 *
 * @return whether the redirection succeeded
 */
bool
CompositeCanvas::redirect_windows()
{
    Window root_return, parent_return;
    Window *children;
    unsigned int nchildren;
    int update_mode;
    unsigned int i;

    update_mode = Options::manual_redirect ? CompositeRedirectManual :
                                             CompositeRedirectAutomatic;

    XGrabServer(xdpy_);

    /*
     * Redirect selected windows.
     */
    if (!Options::window_ids.empty()) {
        for(std::list<Window>::iterator iter = Options::window_ids.begin();
            iter != Options::window_ids.end(); ++iter)
        {
            Window win = *iter;
            /*
             * Redirect window hierarchy to offscreen storage, but provide automatic
             * screen updates (we are not a window/compositing manager!)
             */
            XCompositeRedirectWindow(xdpy_, win, update_mode);

            add_window(win);
        }
    }
    else {
        if (update_mode == CompositeRedirectManual)
            Log::info("Warning: No point using manual redirection when tracking the whole desktop!\n"
                      "         Using automatic redirection instead.\n");


        /*
         * Redirect all windows to offscreen storage, but provide automatic
         * screen updates (we are not a window/compositing manager!)
         */
        XCompositeRedirectSubwindows(xdpy_, root_, CompositeRedirectAutomatic);

        /*
         * Add all the top level windows to the list of windows to watch
         */
        XQueryTree(xdpy_, root_, &root_return, &parent_return,
                   &children, &nchildren);

        for (i = 0; i < nchildren; i++)
            add_window(children[i]);

        XFree(children);
    }

    XUngrabServer(xdpy_);

    return true;
}

/**
 * Handles a XDamage event for a window.
 *
 * @param win the window XID
 */
void
CompositeCanvas::handle_damage(Window win)
{
    /* find window in window list */
    CompositeWindow *comp_win = find_window(win);

    if (!comp_win) {
        Log::debug("Cannot find window 0x%x in list!\n",
                  (unsigned) win);
        return;
    }

    comp_win->handle_damage();
}

/**
 * Handles X events.
 *
 * @return whether the display needs to be redrawn
 */
bool
CompositeCanvas::handle_xevent()
{
    XEvent event;
    bool needs_redraw(false);
    Window window(0);
    Profiler::ScopedSamples xevent_scoped(profiler_xevent_pair_);

    if (Options::idle_redraw && !XPending(xdpy_))
        return true;

    XNextEvent(xdpy_, &event);

    if (event.type == damage_event_ + XDamageNotify) {
        // This is a little ugly, but then so is violating strict aliasing
        // to make the hideous cast to reference this as an XDamageNotifyEvent.
        window = event.xany.window;
        Log::debug("### Damage Win: 0x%x\n", window);
        handle_damage(window);
        needs_redraw = true;
    }
    else if (event.type == ConfigureNotify) {
        window = event.xconfigure.window;
        Log::debug("Configure Win: 0x%x %d x %d\n",
                   window,
                   width_, height_);
        if (window == canvas_) {
            width_ = event.xconfigure.width;
            height_ = event.xconfigure.height;
            reshape(width_, height_);
        }
        else
            update_window(window);
    }
    else if (event.type == Expose) {
        window = event.xexpose.window;
        Log::debug("Expose Win: 0x%x\n", window);
        needs_redraw = true;
    }
    else if (event.type == UnmapNotify) {
        window = event.xunmap.window;
        Log::debug("--- Unmap Win: 0x%x\n", window);
        update_window(window);
        needs_redraw = true;
    }
    else if (event.type == MapNotify) {
        window = event.xmap.window;
        Log::debug("+++ Map Win: 0x%x\n", window);
        update_window(window);
        needs_redraw = true;
    }
    else if (event.type == ReparentNotify) {
        window = event.xreparent.window;
        Window parent = event.xreparent.parent;
        Log::debug("+++ Reparent Win: 0x%x Parent: 0x%x\n", parent);
        if (parent == root_)
            add_window(window);
        else
            remove_window(window);
        needs_redraw = true;
    }
    else if (event.type == CreateNotify) {
        window = event.xcreatewindow.window;
        Log::debug("+++ Add Win: 0x%x Parent: 0x%x\n",
                   window, event.xcreatewindow.parent);
        add_window(window);
    }
    else if (event.type == DestroyNotify) {
        window = event.xdestroywindow.window;
        Log::debug("--- Destroy Win: 0x%x\n", window);
        remove_window(window);
        needs_redraw = true;
    }
    else {
        Log::debug("Unknown event %d\n", event.type);
    }

    return needs_redraw;
}

void
CompositeCanvas::update_all_textures()
{
    for(std::list<CompositeWindow*>::iterator iter = window_list_.begin();
        iter != window_list_.end(); ++iter)
    {
        CompositeWindow *comp_win = *iter;
        if (comp_win->get_texture())
            comp_win->update_texture_from_pixmap(false);
    }
}

void
CompositeCanvas::benchmark_init()
{
    profiler_cycle_pair_.sample_start();
}

void
CompositeCanvas::benchmark_update(bool& iteration_complete)
{
    Profiler::Stats stats;

    iteration_complete = false;

    Profiler &profiler = Profiler::instance();

    profiler_cycle_pair_.sample_end();
    profiler_cycle_pair_.sample_start();

    profiler_cycle_pair_.get_stats(stats);

    // Each iteration is 3 seconds of compositing by the current test object.
    // When we complete an iteration, reset the count and tell the caller.

    if (stats.total >= current_test_duration_) {
        Log::info("FPS: %"PRIu64"\n", (uint64_t)(1000 * stats.nsamples / stats.total));

        for (Profiler::Point p = 0; p < profiler.get_num_points(); p += 2) {
            profiler.get_stats(p, p + 1, stats);
            Log::info("Operation %s => Avg: %f Total: %f Samples: %llu\n",
                      profiler.get_name_by_point(p),
                      stats.average, stats.total, stats.nsamples);
        }

        profiler.reset();
        ++iterations_;
        if (iterations_ < current_test_iterations_) {
            return;
        }
        iterations_ = 0;
        iteration_complete = true;
    }
}

void
CompositeCanvas::load_current_test_options()
{
    std::map<std::string, CompositeTest::Option>::const_iterator iter;

    current_test_duration_ = 0.0;
    current_test_iterations_ = 0;

    iter = current_test_->options().find("duration");
    if (iter != current_test_->options().end()) {
        std::stringstream ss(iter->second.value);
        ss >> current_test_duration_;
        current_test_duration_ *= 1000.0;
    }

    iter = current_test_->options().find("iterations");
    if (iter != current_test_->options().end()) {
        std::stringstream ss(iter->second.value);
        ss >> current_test_iterations_;
    }
}

static void
next_benchmark(std::list<Benchmark*>& benchmarks,
               std::list<Benchmark*>::iterator &benchIt)
{
    benchIt++;
    if (benchIt == benchmarks.end() && Options::run_forever)
        benchIt = benchmarks.begin();
}

/******************
 * Public methods *
 ******************/

/**
 * Initializes the canvas.
 *
 * @return whether the initialization succeeded
 */
bool
CompositeCanvas::init()
{
    /* Open the display */
    xdpy_ = XOpenDisplay(NULL);
    if (!xdpy_)
        return false;

    root_ = RootWindow(xdpy_, DefaultScreen(xdpy_));

    /* Ensure needed extensions (XComposite, Damage etc) are available */
    if (!check_extensions())
        return false;

    XVisualInfo *vis_info = get_canvas_xvisualinfo();
    if (!vis_info)
        return false;

    /* Create the X window we are going to draw on */
    width_ = Options::size;
    height_ = Options::size;
    canvas_ = create_canvas_x_window(glcompbench_version_string.c_str(), 0, 0,
                                     width_, height_,
                                     vis_info);
    XFree(vis_info);

    if (!canvas_)
        return false;

    return true;
}

/**
 * Runs a list of tests.
 *
 * @param benchmarks the list of benchmarks to run
 */
void
CompositeCanvas::run_tests(std::list<Benchmark*> &benchmarks)
{
    if (benchmarks.empty()) {
        Log::error("Test list is empty!\n");
        return;
    }

    if (!make_current()) {
        Log::error("Cannot make current!\n");
        return;
    }

    redirect_windows();

    XMapWindow(xdpy_, canvas_);

    // Initialize all tests so that the benchmarking is not biased by
    // setup time (loading and building shaders, etc.)
    for (std::map<std::string, CompositeTest*>::const_iterator testIt =
            Benchmark::tests().begin();
         testIt != Benchmark::tests().end();
         testIt++)
    {
        CompositeTest* curTest = testIt->second;
        if (!curTest) {
            Log::error("Found a NULL test pointer!\n");
            return;
        }
        curTest->init();
    }

    benchmark_init();

    bool next_iteration(false);

    for (std::list<Benchmark *>::iterator benchIt = benchmarks.begin();
         benchIt != benchmarks.end();
         next_benchmark(benchmarks, benchIt))
    {
        Benchmark *benchmark = *benchIt;
        current_test_ = &benchmark->setup_test();

        if (!current_test_->name().empty()) {
            load_current_test_options();
            reshape(width_, height_);
            Log::info("Running test: '%s'\n", current_test_->name().c_str());

            while (true) {
                if (handle_xevent()) {
                    if (Options::force_tex_update)
                        update_all_textures();

                    if (Options::draw) {
                        profiler_draw_pair_.sample_start();
                        current_test_->draw(this->window_list_);
                        profiler_draw_pair_.sample_end();
                        profiler_screen_update_pair_.sample_start();
                        update_screen();
                        profiler_screen_update_pair_.sample_end();
                    }

                    benchmark_update(next_iteration);
                    if (next_iteration) {
                        break;
                    }

                }
            }
        }

        benchmark->teardown_test();
    }
}

