/*
 * Copyright © 2011 Linaro Limited
 *
 * This file is part of glcompbench.
 *
 * glcompbench is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * glcompbench is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with glcompbench.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authors:
 *  Alexandros Frantzis <alexandros.frantzis@linaro.org>
 *  Jesse Barker <jesse.barker@linaro.org>
 */

#include "composite-test.h"
#include "options.h"
#include "log.h"

using std::string;
using std::map;

const string CompositeTestSimpleBase::model_view_matrix_name_("modelview");
const string CompositeTestSimpleBase::projection_matrix_name_("projection");
const string CompositeTestSimpleBase::texture0_name_("Texture0");
const string CompositeTestSimpleBase::texcoord_name_("texcoord");
const string CompositeTestSimpleBase::position_name_("position");

CompositeTestSimpleBase::CompositeTestSimpleBase(const std::string& test_name,
                                                 const std::string& vs_filename,
                                                 const std::string& fs_filename)
    : CompositeTest(test_name), vs_filename_(vs_filename),
      fs_filename_(fs_filename)
{
    options_["use-vbo"] = CompositeTest::Option("use-vbo", "true",
                                                "Whether to use VBOs for rendering [true,false]");
}

void
CompositeTestSimpleBase::init()
{
    // Initialize shader sources from input files
    if (!gotSource(vs_filename_, vertex_shader_))
    {
        Log::error("Failed to get vertex shader source for test\n");
        return;
    }
    if (!gotSource(fs_filename_, fragment_shader_))
    {
        Log::error("Failed to get fragment shader source for test\n");
        return;
    }

    program_.init();
    if (!program_.valid())
    {
        Log::error("No valid program for compositing (reason: %s)\n",
                   program_.errorMessage().c_str());
        return;
    }
    program_.addShader(GL_VERTEX_SHADER, vertex_shader_);
    if (!program_.valid())
    {
        Log::error("Failed to add vertex shader to compositing program (reason: %s)\n",
                   program_.errorMessage().c_str());
        return;
    }
    program_.addShader(GL_FRAGMENT_SHADER, fragment_shader_);
    if (!program_.valid())
    {
        Log::error("Failed to add fragment shader to compositing program (reason: %s)\n",
                   program_.errorMessage().c_str());
        return;
    }
    program_.build();
    if (!program_.ready())
    {
        Log::error("Failed to build compositing program (reason: %s)\n",
                   program_.errorMessage().c_str());
        return;
    }
    vertexIndex_ = program_.getAttribIndex(position_name_);
    texcoordIndex_ = program_.getAttribIndex(texcoord_name_);

    // Initialize our vertex buffer object
    using LibMatrix::vec2;
    using LibMatrix::vec3;
    vboData_.addVertex(vec3(-1, -1, 0));
    vboData_.addVertex(vec3( 1, -1, 0));
    vboData_.addVertex(vec3( 1,  1, 0));
    vboData_.addVertex(vec3(-1,  1, 0));

    vboData_.addIndex(0);
    vboData_.addIndex(1);
    vboData_.addIndex(2);
    vboData_.addIndex(2);
    vboData_.addIndex(3);
    vboData_.addIndex(0);

    vboData_.addTexCoord(vec2(0, 1));
    vboData_.addTexCoord(vec2(1, 1));
    vboData_.addTexCoord(vec2(1, 0));
    vboData_.addTexCoord(vec2(0, 0));

    vboData_.genBufferObject();
}

/**
 * Prepares the test for a test run.
 */
void
CompositeTestSimpleBase::prepare_for_run()
{
    vboData_.useBufferObject(options_["use-vbo"].value == "true");

    program_.start();
    int texUnit(0);
    program_.loadUniformScalar(texUnit, texture0_name_);

    glClearColor(0.1, 0.1, 0.3, 1.0);

    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
}

void
CompositeTestSimpleBase::reshape(int width, int height)
{
    float ar = static_cast<float>(width) / static_cast<float>(height);

    /* Update the projection matrix */
    projection_matrix.loadIdentity();
    projection_matrix.perspective(22.6, ar, 5.0, 60.0);

    /* Update the modelview matrix */
    model_view_matrix.loadIdentity();
    model_view_matrix.translate(0.0, 0.0, -10.0);

    /* Set the viewport */
    glViewport(0, 0, width, height);
}

void
CompositeTestSimpleBase::cleanup()
{
    program_.stop();
}
