/*
 * Copyright © 2011 Linaro Limited
 *
 * This file is part of glcompbench.
 *
 * glcompbench is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * glcompbench is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with glcompbench.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authors:
 *  Alexandros Frantzis <alexandros.frantzis@linaro.org>
 *  Jesse Barker <jesse.barker@linaro.org>
 */

#include "composite-test.h"
#include "options.h"
#include "log.h"

using std::string;
using LibMatrix::vec2;
using LibMatrix::vec3;

const string CompositeTestSimpleBrick::lightPosName_("LightPosition");
const string CompositeTestSimpleBrick::brickSizeName_("BrickSize");
const string CompositeTestSimpleBrick::brickPctName_("BrickPct");
const vec2 CompositeTestSimpleBrick::brickSize_(0.30, 0.15);
const vec2 CompositeTestSimpleBrick::brickPct_(0.90, 0.85);

void
CompositeTestSimpleBrick::prepare_for_run()
{
    CompositeTestSimpleBase::prepare_for_run();

    lightPos_ = LibMatrix::vec4(0.0, 1.0, 1.0, 0.0);

    program_.loadUniformVector(brickSize_, brickSizeName_);
    program_.loadUniformVector(brickPct_, brickPctName_);
    program_.loadUniformVector(lightPos_, lightPosName_);
}

void
CompositeTestSimpleBrick::draw(std::list<CompositeWindow *> &window_list)
{
    vboData_.bind();
    glActiveTexture(GL_TEXTURE0);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    /* Set up the position of the attributes in the vertex array */
    glVertexAttribPointer(vertexIndex_, 3, GL_FLOAT, GL_FALSE, 0, vboData_.vertexOffset());
    glVertexAttribPointer(texcoordIndex_, 2, GL_FLOAT, GL_FALSE, 0, vboData_.texcoordOffset());

    /* Enable the attributes */
    glEnableVertexAttribArray(vertexIndex_);
    glEnableVertexAttribArray(texcoordIndex_);

    program_.loadUniformMatrix(projection_matrix.getCurrent(),
                               projection_matrix_name_);

    /* Find out how many windows are visible and calculate the angular step */
    GLuint visible_windows(num_visible_windows(window_list));
    GLfloat angular_step(2 * M_PI / visible_windows);

    /* Draw the windows in a circle using the calculated angular step */
    GLint i(0);
    for(std::list<CompositeWindow*>::iterator iter = window_list.begin();
        iter != window_list.end(); ++iter)
    {
        CompositeWindow *comp_win = *iter;
        GLuint tex = comp_win->get_texture();
        if (tex) {
            model_view_matrix.push();
            model_view_matrix.translate(cos(angular_step * i),
                                        sin(angular_step * i), 0);

            /* Load shader ModelView uniform */
            program_.loadUniformMatrix(model_view_matrix.getCurrent(),
                                       model_view_matrix_name_);

            Log::debug("Drawing Win: 0x%x Tex: 0x%x\n",
                       comp_win->get_xwindow(), comp_win->get_texture());

            glBindTexture(GL_TEXTURE_2D, tex);
            vboData_.draw();
            model_view_matrix.pop();
            ++i;
        }
    }

    /* Disable the attributes */
    glDisableVertexAttribArray(vertexIndex_);
    glDisableVertexAttribArray(texcoordIndex_);

    vboData_.unbind();

    // Pan the light around the scene.
    float x(lightPos_.x());
    float y(lightPos_.y());
    if (x >= 1.0) { x = 1.0; y -= 0.1; }
    if (y >= 1.0) { y = 1.0; x += 0.1; }
    if (y <= 0.0) { y = 0.0; x -= 0.1; }
    if (x <= 0.0) { x = 0.0; y += 0.1; }
    lightPos_.x(x);
    lightPos_.y(y);
    program_.loadUniformVector(lightPos_, lightPosName_);
}
