/*
 * Copyright © 2011 Linaro Limited
 *
 * This file is part of glcompbench.
 *
 * glcompbench is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * glcompbench is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with glcompbench.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authors:
 *  Alexandros Frantzis <alexandros.frantzis@linaro.org>
 *  Jesse Barker <jesse.barker@linaro.org>
 */

#include <X11/Xlib.h>
#include <X11/extensions/Xcomposite.h>
#include <X11/extensions/Xdamage.h>

#include "composite-window.h"
#include "log.h"
#include "profiler.h"

CompositeWindow::~CompositeWindow()
{
    if (pix_)
        XFreePixmap(xdpy_, pix_);

    if (damage_)
        XDamageDestroy(xdpy_, damage_);

    if (tex_)
        glDeleteTextures(1, &tex_);
}

/**
 * Updates a CompositeWindow.
 */
void
CompositeWindow::update()
{
    int x, y;
    unsigned int width, height, border;
    Window root;
    bool update_texture = false;
    Pixmap pix_to_free = None;

    Log::debug("Composite: Updating window 0x%x\n", (unsigned)win_);

    XGetGeometry(xdpy_, win_, &root,
                 &x, &y, &width, &height,
                 &border, &depth_);

    /* If the window has been resized the backing pixmap is not valid */
    if (width != width_ || height != height_) {
        Log::debug("    => Resize detected, updating pixmap (old W: %u H: %u)\n",
                  width_, height_);
        width_ = width;
        height_ = height;
        pix_to_free = pix_;
        pix_ = None;
    }

    XGetWindowAttributes(xdpy_, win_, &attr_);

    Log::debug("    => W: %u H: %u Mapped: 0x%x\n",
              width_, height_, attr_.map_state);

    /* Update pixmap */
    if (!pix_ && attr_.map_state != 0) {
        pix_ = XCompositeNameWindowPixmap(xdpy_, win_);
        Log::debug("    => Pixmap <= 0x%x\n", (unsigned) pix_);
        update_texture = true;
    }
    else if (pix_ && attr_.map_state == 0) {
        Log::debug("    => Destroying Pixmap 0x%x\n", (unsigned) pix_);
        pix_to_free = pix_;
        pix_ = None;
    }

    Log::debug("    => Damage State <= 0x%x\n", (unsigned) damage_);
    /* Update damage object */
    if (!damage_ && attr_.map_state != 0) {
        damage_ = XDamageCreate(xdpy_, win_, XDamageReportNonEmpty);
        Log::debug("    => Damage <= 0x%x\n", (unsigned) damage_);
    }
    else if (damage_ && attr_.map_state == 0) {
        Log::debug("    => Destroying damage 0x%x\n", (unsigned) damage_);
        XDamageDestroy(xdpy_, damage_);
        damage_ = None;
    }

    /* Create or delete texture if needed */
    if (!tex_ && attr_.map_state != 0) {
        glGenTextures(1, &tex_);

        glBindTexture(GL_TEXTURE_2D, tex_);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

        Log::debug("    => texture <= 0x%x\n", (unsigned) tex_);
        update_texture = true;
    }
    else if (tex_ && attr_.map_state == 0) {
        Log::debug("    => Deleting texture 0x%x\n", (unsigned) tex_);
        glDeleteTextures(1, &tex_);
        tex_ = 0;
    }

    if (update_texture)
        update_texture_from_pixmap(true);

    /*
     * We have to free the pixmap after calling update_texture_from_pixmap(),
     * because glXDestroyPixmap() in Mesa (used in the GLX implementation of
     * update_texture_from_pixmap()) fails with BadDrawable when the X Pixmap
     * associated with the GLX pixmap has been previously freed.
     */
    if (pix_to_free)
        XFreePixmap(xdpy_, pix_to_free);
}

Window
CompositeWindow::get_xwindow()
{
    return win_;
}

GLuint
CompositeWindow::get_texture()
{
    return tex_;
}

void
CompositeWindow::handle_damage()
{
    if (damage_)
        XDamageSubtract(xdpy_, damage_, None, None);

    update_texture_from_pixmap(false);
}
