/*
 * Copyright © 2011 Linaro Limited
 *
 * This file is part of glcompbench.
 *
 * glcompbench is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * glcompbench is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with glcompbench.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authors:
 *  Alexandros Frantzis <alexandros.frantzis@linaro.org>
 *  Jesse Barker <jesse.barker@linaro.org>
 */

#include <list>

#ifdef USE_EGL
#include "composite-canvas-egl.h"
#elif USE_GLX
#include "composite-canvas-glx.h"
#endif

#include "composite-test.h"
#include "benchmark.h"
#include "options.h"
#include "log.h"

static const char *default_benchmarks[] = {
    "default",
    "brick",
    NULL
};

static void
list_tests()
{
    const map<string, CompositeTest *> &tests = Benchmark::tests();

    for (map<string, CompositeTest *>::const_iterator test_iter = tests.begin();
         test_iter != tests.end();
         test_iter++)
    {
        CompositeTest *test = test_iter->second;
        if (test->name().empty())
            continue;

        Log::info("[Test] %s\n", test->name().c_str());

        const map<string, CompositeTest::Option> &options = test->options();

        for (map<string, CompositeTest::Option>::const_iterator opt_iter = options.begin();
             opt_iter != options.end();
             opt_iter++)
        {
            const CompositeTest::Option &opt = opt_iter->second;
            Log::info("  [Option] %s\n"
                      "    Description  : %s\n"
                      "    Default Value: %s\n",
                      opt.name.c_str(), 
                      opt.description.c_str(),
                      opt.default_value.c_str());
        }
    }
}

static void
add_default_benchmarks(std::list<Benchmark*> &benchmarks)
{
    for (const char **s = default_benchmarks; *s != NULL; s++)
        benchmarks.push_back(new Benchmark(*s));
}

static void
add_custom_benchmarks(std::list<Benchmark *> &benchmarks)
{
    for (std::list<string>::const_iterator iter = Options::benchmarks.begin();
         iter != Options::benchmarks.end();
         iter++)
    {
        Log::debug("Adding benchmark '%s'\n", iter->c_str());
        benchmarks.push_back(new Benchmark(*iter));
    }
}

int
main(int argc, char **argv)
{
#ifdef USE_EGL
    CompositeCanvasEGL canvas;
#elif USE_GLX
    CompositeCanvasGLX canvas;
#endif

    std::list<Benchmark*> benchmarks;

    if (!Options::parse_args(argc, argv))
        return 1;

    if (Options::show_help) {
        Options::print_help();
        return 0;
    }

    Benchmark::register_test(*new CompositeTestDefaultOptions());
    Benchmark::register_test(*new CompositeTestSimpleDefault());
    Benchmark::register_test(*new CompositeTestSimpleBrick());

    if (Options::list_tests) {
        list_tests();
        return 0;
    }

    // Add the benchmarks to run
    if (Options::benchmarks.empty())
        add_default_benchmarks(benchmarks);
    else
        add_custom_benchmarks(benchmarks);

    canvas.init();

    Log::info("=======================================================\n");
    Log::info("    glcompbench %s\n", GLCOMPBENCH_VERSION);
    Log::info("=======================================================\n");
    Log::info("%s", canvas.info_string().c_str());
    Log::info("=======================================================\n");

    canvas.run_tests(benchmarks);

    return 0;
}
