/*
 * Copyright © 2011 Linaro Limited
 *
 * This file is part of glcompbench.
 *
 * glcompbench is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * glcompbench is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with glcompbench.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authors:
 *  Alexandros Frantzis <alexandros.frantzis@linaro.org>
 *  Jesse Barker <jesse.barker@linaro.org>
 */

#define __STDC_LIMIT_MACROS
#include <sys/time.h>
#include <stdint.h>
#include <string>
#include <vector>

#include "profiler.h"

struct ProfilerPoint
{
    ProfilerPoint(const char *name_):
        name(name_), prev(0), next(0)
    {
        reset();
    }

    void reset()
    {
        last_sample = 0;
        nsamples = 0;
        low = UINT64_MAX;
        high = 0;
        total = 0;
        average = 0.0;
    }

    std::string name;
    ProfilerPoint *prev;
    ProfilerPoint *next;
    uint64_t last_sample;
    uint64_t nsamples;
    uint64_t low;
    uint64_t high;
    uint64_t total;
    double average;
};

Profiler::~Profiler()
{
    for(std::vector<ProfilerPoint*>::iterator iter = points_.begin();
        iter != points_.end(); ++iter)
    {
        delete (*iter);
    }
}

/**
 * Registers a profiling point.
 *
 * The handles of Profiler points are of a signed numerical type.
 * Each succesive point registration is guaranteed to produce a
 * handle that is +1 from the last one.
 *
 * @param name the human readable name of the point
 *
 * @return the unique handle used to reference this point
 */
Profiler::Point
Profiler::register_point(const char *name)
{
    ProfilerPoint *point = new ProfilerPoint(name);

    points_.push_back(point);

    return points_.size() - 1;
}

/**
 * Gets the handle of a profiling point given its name.
 *
 * @param name the name of the profiling by
 *
 * @return the point handle
 */
Profiler::Point
Profiler::get_point_by_name(const char *name)
{
    for(std::vector<ProfilerPoint*>::iterator iter = points_.begin();
        iter != points_.end(); ++iter)
    {
        if ((*iter)->name == name)
            return std::distance(points_.begin(), iter);
    }

    return -1;
}

/**
 * Gets the human readable name of a profiling point.
 *
 * @param p the point handle
 *
 * @return name of the point
 */
const char *
Profiler::get_name_by_point(Profiler::Point p)
{
    if (!is_valid_point(p))
        return NULL;

    return points_[p]->name.c_str();
}

/**
 * Connects two profiling points.
 *
 * @param p1 the first point
 * @param p2 the second (successor) point
 *
 * @return whether the operation succeeded
 */
bool
Profiler::connect(Profiler::Point p1, Profiler::Point p2)
{
    if (!is_valid_point(p1) || !is_valid_point(p2))
        return false;

    points_[p1]->next = points_[p2];
    points_[p2]->prev = points_[p1];

    return true;
}

/**
 * Gets the first point that has been sampled and is a predecessor of the
 * supplied point.
 *
 * @param point the point to start searching at
 *
 * @return the found point or NULL
 */
static ProfilerPoint *
get_prev_sampled_point(ProfilerPoint *point)
{
    ProfilerPoint *cur = point->prev;
    while (cur != NULL && cur->last_sample == 0 && cur != point)
        cur = cur->prev;

    if (cur != NULL && cur->last_sample != 0)
        return cur;
    else
        return NULL;
}

/**
 * Samples a profiling point.
 *
 * @param p the profiling point handle
 *
 * @return whether the operation succeeded
 */
bool
Profiler::sample_point(Profiler::Point p)
{
    if (!is_valid_point(p))
        return false;

    ProfilerPoint *point = points_[p];
    ProfilerPoint *prev = get_prev_sampled_point(point);
    uint64_t now = get_timestamp_us();

    if (prev != NULL) {
        uint64_t diff = now - point->prev->last_sample;
        point->average = (point->nsamples * point->average + diff) /
                         (point->nsamples + 1);
        point->total += diff;
        if (diff < point->low)
            point->low = diff;
        if (diff > point->high)
            point->high = diff;
        point->nsamples++;
    }

    point->last_sample = now;

    return true;
}

/**
 * Gets the statistics for the run between two points.
 *
 * @param p1 the first point
 * @param p2 the second point (may be the same as the first for a complete loop)
 * @param stats the struct to fill
 *
 * @return whether the operation succeeded
 */
bool
Profiler::get_stats(Profiler::Point p1,
                    Profiler::Point p2,
                    Stats &stats)
{
    if (!is_valid_point(p1) || !is_valid_point(p2))
        return false;

    ProfilerPoint *cur = points_[p1]->next;
    ProfilerPoint *end = points_[p2];

    /* Reset statistics */
    stats.nsamples = 0;
    stats.low = 0.0;
    stats.high = 0.0;
    stats.total = 0.0;
    stats.average = 0.0;

    /* Go from p1 to p2, gathering stats */
    while (cur != NULL) {
        /*
         * We store values as microseconds internally, but provide them as
         * milliseconds
         */
        stats.average += cur->average / 1000.0;
        stats.total += cur->total / 1000.0;
        stats.low += cur->low / 1000.0;
        stats.high += cur->high / 1000.0;
        stats.nsamples = std::max(stats.nsamples, cur->nsamples);
        if (cur == end)
            break;
        cur = cur->next;
    }

    return (cur == end);
}

/**
 * Resets the benchmarking info of a profiling point.
 *
 * @param p the point to reset
 */
void
Profiler::reset_point(Profiler::Point p)
{
    if (!is_valid_point(p))
        return;

    points_[p]->reset();
}

/**
 * Resets the benchmarking info for all points in the profiler.
 */
void
Profiler::reset(void)
{
    for(size_t i = 0; i < points_.size(); i++)
        points_[i]->reset();
}

/**
 * Checks whether a profiling point handle is valid.
 *
 * @param p the profiling handle
 *
 * @return whether the profiling point handle is valid
 */
bool
Profiler::is_valid_point(Profiler::Point p)
{
    return (p >= 0 && p < (Profiler::Point) points_.size());
}

/**
 * Gets the current timestamp.
 *
 * @return the current timestam in microseconds
 */
uint64_t
Profiler::get_timestamp_us()
{
    struct timeval tv;
    gettimeofday(&tv, NULL);
    uint64_t now = static_cast<uint64_t>(tv.tv_sec) * 1000000 +
                   static_cast<double>(tv.tv_usec);
    return now;
}

