/*
 * Copyright © 2011 Linaro Limited
 *
 * This file is part of glcompbench.
 *
 * glcompbench is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * glcompbench is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with glcompbench.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authors:
 *  Alexandros Frantzis <alexandros.frantzis@linaro.org>
 *  Jesse Barker <jesse.barker@linaro.org>
 */

#ifndef PROFILER_H_
#define PROFILER_H_

#define __STDC_LIMIT_MACROS
#include <stdint.h>
#include <vector>

using namespace std;

struct ProfilerPoint;

class Profiler
{
public:
    typedef int Point;

    struct Stats
    {
        uint64_t nsamples;
        double low;
        double high;
        double total;
        double average;
    };

    class PointPair
    {
    public:
        PointPair(Profiler &profiler, const char *name) :
            profiler_(profiler)
        {
            init(name);
        }

        PointPair(const char *name)
            : profiler_(Profiler::instance())
        {
            init(name);
        }

        Profiler::Point get_start()
        {
            return start_;
        }

        Profiler::Point get_end()
        {
            return end_;
        }

        Profiler &get_profiler()
        {
            return profiler_;
        }

        void sample_start()
        {
            profiler_.sample_point(start_);
        }

        void sample_end()
        {
            profiler_.sample_point(end_);
        }

        bool get_stats(Profiler::Stats &stats)
        {
            return profiler_.get_stats(start_, end_, stats);
        }

    private:
        void init(const char *name)
        {
            start_ = profiler_.register_point(name);
            end_ = profiler_.register_point(name);
            profiler_.connect(start_, end_);
        }

        Profiler &profiler_;
        Profiler::Point start_;
        Profiler::Point end_;
    };

    class ScopedSamples
    {
    public:
        ScopedSamples(PointPair &pair) :
            pair_(pair)
        {
            pair_.sample_start();
        }

        ~ScopedSamples()
        {
            pair_.sample_end();
        }

    private:
        Profiler::PointPair &pair_;
    };


    Profiler() {}
    ~Profiler();

    Point register_point(const char *name);
    Point get_point_by_name(const char *name);
    const char *get_name_by_point(Point p);
    bool connect(Point p1, Point p2);
    bool sample_point(Point p);
    bool get_stats(Point p1, Point p2, Stats &stats);
    void reset_point(Point p);
    void reset(void);
    Point get_num_points() { return points_.size(); }

    /**
     * Gets the profiler singleton
     */
    static Profiler& instance()
    {
        static Profiler profiler;
        return profiler;
    }

private:
    bool is_valid_point(Point p);
    uint64_t get_timestamp_us();
    std::vector<ProfilerPoint *> points_;
};


#endif /* PROFILER_H_ */
