//
// Copyright © 2011 Linaro Limited
//
// This file is part of glcompbench.
//
// glcompbench is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// glcompbench is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with glcompbench.  If not, see <http://www.gnu.org/licenses/>.
//
// Authors:
//  Alexandros Frantzis <alexandros.frantzis@linaro.org>
//  Jesse Barker <jesse.barker@linaro.org>
//
#include <iostream>
#include "vbo.h"
#include "gl-headers.h"

using std::string;
using std::vector;
using std::ifstream;
using LibMatrix::vec2;
using LibMatrix::vec3;

void
VertexList::getVertices(std::ifstream& in)
{
    unsigned short numVertices(0);
    in.read(reinterpret_cast<char*>(&numVertices), sizeof(unsigned short));
    vertexVec_.reserve(numVertices);
    for (unsigned int i = 0; i < numVertices; i++)
    {
        float x, y, z;
        in.read(reinterpret_cast<char*>(&x), sizeof(float));
        in.read(reinterpret_cast<char*>(&y), sizeof(float));
        in.read(reinterpret_cast<char*>(&z), sizeof(float));
        vertexVec_.push_back(vec3(x, y, z));
    }
}

void
FaceList::getFaces(std::ifstream& in)
{
    unsigned short numFaces(0);
    in.read(reinterpret_cast<char*>(&numFaces), sizeof(unsigned short));
    indexVec_.reserve(numFaces);
    for (unsigned int i = 0; i < numFaces; i++)
    {
        // A face is 3 indices.
        unsigned short idx(0);
        in.read(reinterpret_cast<char*>(&idx), sizeof(unsigned short));
        indexVec_.push_back(idx);
        in.read(reinterpret_cast<char*>(&idx), sizeof(unsigned short));
        indexVec_.push_back(idx);
        in.read(reinterpret_cast<char*>(&idx), sizeof(unsigned short));
        indexVec_.push_back(idx);
        // Get the face flags to advance the stream pointer.
        in.read(reinterpret_cast<char*>(&idx), sizeof(unsigned short));
    }
}

void
TexCoordList::getTexCoords(std::ifstream& in)
{
    unsigned short numTexCoords(0);
    in.read(reinterpret_cast<char*>(&numTexCoords), sizeof(unsigned short));
    texcoordVec_.reserve(numTexCoords);
    for (unsigned int i = 0; i < numTexCoords; i++)
    {
        float s, t;
        in.read(reinterpret_cast<char*>(&s), sizeof(float));
        in.read(reinterpret_cast<char*>(&t), sizeof(float));
        texcoordVec_.push_back(vec2(s, t));
    }
}

VertexBufferData::VertexBufferData() :
    useBufferObjects_(false),
    isMutable_(false)
{
    bufferObjects_[0] = bufferObjects_[1] = 0;
}

VertexBufferData::~VertexBufferData()
{
    // If we didn't gen them, they're 0 and we can still call delete.
    glDeleteBuffers(2, &bufferObjects_[0]);
}

bool
VertexBufferData::getFrom3dsFile(const string& filename)
{
    ifstream inputFile(filename.c_str());
    if (!inputFile)
    {
        std::cerr << "Failed to open \"%s\"" << filename << std::endl;
        return false;
    }

    while (!inputFile.eof())
    {
        // Read the chunk identifier
        unsigned short chunkID(0);
        inputFile.read(reinterpret_cast<char*>(&chunkID), sizeof(unsigned short));
        // Read the chunk length
        unsigned int chunkSize(0);
        inputFile.read(reinterpret_cast<char*>(&chunkSize), sizeof(unsigned int));

        string objectName;
        unsigned int i(0);
        char curChar(0);
        switch (chunkID)
        {
            case 0x4d4d:
                // Main chunk
                break;
            case 0x3d3d:
                // 3D Editor chunk
                break;
            case 0x4000:
                // Object block
                // Name is max 20 characters.
                do
                {
                    inputFile.get(curChar);
                    objectName.push_back(curChar);
                    i++;
                } while (curChar != '\0' && i < 20);
                break;
            case 0x4100:
                // Triangular mesh
                break;
            case 0x4110:
                // Vertices list
                vertexData_.getVertices(inputFile);
                break;
            case 0x4120:
                // Faces description
                indexData_.getFaces(inputFile);
                break;
            case 0x4140:
                // Mapping Coordinate list
                texcoordData_.getTexCoords(inputFile);
                break;
            default:
                // We're already 6 bytes into the chunk after having read the
                // ID and length, so subtract it from the length to get the 
                // seek offset.
                inputFile.seekg(chunkSize-6, std::ios::cur);
                break;
        }
    }

    return true;
}

void
VertexBufferData::setMutable(bool isMutable)
{
    isMutable_ = isMutable;

    if (!isMutable_)
    {
        // Regardless of how we've setup the vertex data, we can setup the map of
        // how that data will be laid out in the buffer object.  Default to 
        // buffer object layout ("linear" map).
        dataMap_.indexOffset = 0;
        dataMap_.indexSize = indexData_.count() * sizeof(unsigned short);
        dataMap_.vertexOffset = 0;
        dataMap_.vertexSize = vertexData_.count() * sizeof(vec3);
        dataMap_.totalSize = dataMap_.vertexSize;
        dataMap_.texcoordOffset = dataMap_.vertexOffset + dataMap_.vertexSize;
        dataMap_.texcoordSize = texcoordData_.count() * sizeof(vec2);
        dataMap_.totalSize += dataMap_.texcoordSize;

        // If we are not using buffer objects, the sizes don't change, only the
        // offsets (they become real pointers to the data in client memory).
        if (!useBufferObjects_)
        {
            dataMap_.indexOffset = reinterpret_cast<size_t>(indexData_.data());
            dataMap_.vertexOffset = reinterpret_cast<size_t>(vertexData_.data());
            dataMap_.texcoordOffset = reinterpret_cast<size_t>(texcoordData_.data());
        }
    }
}

void
VertexBufferData::useBufferObject(bool useBO)
{
    useBufferObjects_ = useBO;
    // Freeze the data and update the meta-data we use to feed the
    // implementation.
    setMutable(false);
}

void
VertexBufferData::genBufferObject()
{
    // Mark us as using the buffer objects.
    useBufferObject(true);

    // We need 2 buffers for our work here.  One for the vertex data.
    // and one for the index data.
    glGenBuffers(2, &bufferObjects_[0]);

    // Bind our buffer objects so we can initialize their state.
    bind();

    // First, setup the vertex data by allocating its data store, and filling 
    // it in with our vertex and texcoord data.
    glBufferData(GL_ARRAY_BUFFER, dataMap_.totalSize, 0, GL_STATIC_DRAW);
    glBufferSubData(GL_ARRAY_BUFFER, dataMap_.vertexOffset, dataMap_.vertexSize,
                    vertexData_.data());
    glBufferSubData(GL_ARRAY_BUFFER, dataMap_.texcoordOffset, dataMap_.texcoordSize,
                    texcoordData_.data());

    // Now repeat for our index data.
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, dataMap_.indexSize,
                 indexData_.data(), GL_STATIC_DRAW);

    // Unbind them so we don't get any unexpected changes to the buffer object
    // state.
    unbind();
}

void
VertexBufferData::bind()
{
    if (useBufferObjects_)
    {
        glBindBuffer(GL_ARRAY_BUFFER, bufferObjects_[0]);
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, bufferObjects_[1]);
    }
}

void
VertexBufferData::unbind()
{
    if (useBufferObjects_)
    {
        glBindBuffer(GL_ARRAY_BUFFER, 0);
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
    }
}

void
VertexBufferData::draw()
{
    glDrawElements(GL_TRIANGLES, indexData_.count(), GL_UNSIGNED_SHORT,
        indexOffset());
}
