//
// Copyright © 2011 Linaro Limited
//
// This file is part of glcompbench.
//
// glcompbench is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// glcompbench is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with glcompbench.  If not, see <http://www.gnu.org/licenses/>.
//
// Authors:
//  Alexandros Frantzis <alexandros.frantzis@linaro.org>
//  Jesse Barker <jesse.barker@linaro.org>
//
#ifndef VBO_H_
#define VBO_H_

#include <fstream>
#include <vector>
#include "vec.h"

class VertexList
{
public:
    VertexList() {}
    ~VertexList() {}
    void getVertices(std::ifstream& in);
    void addVertex(const LibMatrix::vec3& v) { vertexVec_.push_back(v); }
    unsigned int count() const { return vertexVec_.size(); }
    const LibMatrix::vec3* data() const { return vertexVec_.data(); }
private:
    std::vector<LibMatrix::vec3> vertexVec_;
};

class FaceList
{
public:
    FaceList() {}
    ~FaceList() {}
    void getFaces(std::ifstream& in);
    void addIndex(const unsigned short i) { indexVec_.push_back(i); }
    unsigned int count() const { return indexVec_.size(); }
    const unsigned short* data() { return indexVec_.data(); }
private:
    std::vector<unsigned short> indexVec_;
};

class TexCoordList
{
public:
    TexCoordList() {}
    ~TexCoordList() {}
    void getTexCoords(std::ifstream& in);
    void addTexCoord(const LibMatrix::vec2& t) { texcoordVec_.push_back(t); }
    unsigned int count() const { return texcoordVec_.size(); }
    const LibMatrix::vec2* data() const { return texcoordVec_.data(); }
private:
    std::vector<LibMatrix::vec2> texcoordVec_;
};


class VertexBufferData
{
public:
    VertexBufferData();
    ~VertexBufferData();

    // Adds a single vertex to the array.
    void addVertex(const LibMatrix::vec3& v) { vertexData_.addVertex(v); }
    // Adds a single index to the array.
    void addIndex(const unsigned short i) { indexData_.addIndex(i); }
    // Adds a single texture coordinate to the array.
    void addTexCoord(const LibMatrix::vec2& t) { texcoordData_.addTexCoord(t); }
    // Utility function to extract vertex, index and texture coordinate data
    // from a .3ds file.  "filename" is a fully-qualified path to the file.
    bool getFrom3dsFile(const std::string& filename);
    // As the implementation needs access to our data at certain times, we
    // require the caller to signal when they will change the data.  Calling
    // setMutable(true) indicates the data may be updated, and calling it with
    // "false" means the data will not be changed and we may rely on the current
    // state.
    void setMutable(bool isMutable);
    // Creates appropriate vertex buffer objects from the current state of the
    // vertex array data.  Implicitly sets the data to be immutable (see 
    // "setMutable()" member) but the buffer objects are left unbound to
    // prevent corruption by other code sharing the same context.
    void genBufferObject();
    // If 'useBO' == true, data will be referenced from the buffer objects.
    // If 'useBO' == false, data will be referenced from client arrays.
    // Affects the behavior of 'setMutable()' as well as all of the following
    // functions (i.e. this acts as a switch so all other calling code can be
    // identical).
    void useBufferObject(bool useBO);
    // Setup for drawing.
    void bind();
    // Done drawing.
    void unbind();
    // Draw the object(s) described by the current data.
    void draw();
    // Used to specifiy pointers to glVertexAttribPointer(), etc.
    const void* texcoordOffset() const { return reinterpret_cast<const void*>(dataMap_.texcoordOffset); }
    const void* vertexOffset() const { return reinterpret_cast<const void*>(dataMap_.vertexOffset); }
    const void* indexOffset() const { return reinterpret_cast<const void*>(dataMap_.indexOffset); }
private:
    // A simple map so we know where each section of our data starts within
    // our vertex buffer object.
    struct BufferDataMap
    {
        size_t indexOffset;
        size_t indexSize;
        size_t vertexOffset;
        size_t vertexSize;
        size_t texcoordOffset;
        size_t texcoordSize;
        size_t totalSize;
    } dataMap_;
    VertexList vertexData_;
    TexCoordList texcoordData_;
    FaceList indexData_;
    // One for vertex array, one for index array
    unsigned int bufferObjects_[2];
    bool useBufferObjects_;
    bool isMutable_;
};

#endif // VBO_H_
